import re
from mimetypes import types_map
types_set = set(types_map.values())
from funmap import funmap

def _is_content_type(x):
    return x.count('/') == 1
    return x in types_set or \
        x.count('/') == 1 and x.split('/', 1)[1].lower().startswith('x-')

@funmap('original', 'parsed')
def ContentType(content_type):
    original = content_type
    if isinstance(content_type, str):
        if content_type in types_map:
            parsed = types_map[content_type]
        else:
            y = re.split(r'[ ;]', content_type)[0]
            if _is_content_type(y):
                parsed = y
            else:
                raise ValueError('Bad content type: %s' % content_type)
    else:
        raise TypeError('Content-type must be str, not %s' % type(content_type).__name__)
    return {'original': original, 'parsed': parsed}



import logging
from inspect import getmembers
from itertools import chain
from triedict import triedict
from collections import Mapping
from .function import Function
from .config import Configuration
logger = logging.getLogger(__name__)

class Program(Mapping):
    def __init__(self, function, *configurations, name=None):
        '''
        :param functions: Base function or functions
        :param Configuration configurations: Configuration files and other
            configuration sources
        :param name: Force the program to have a particular name.
        '''
        if name is None and hasattr(function, '__call__'):
            name = function.__name__
        self.name = name
        self._function = function
        self._configurations = tuple(map(Configuration, configurations))
        self._d = triedict(_sub(tuple(), function))

    def subset(self, prefix):
        for section, function in self.items():
            if section[:len(prefix)] == prefix:
                yield section

    def __repr__(self):
        tpl = '%(class)s(%(function)s, %(configuration)s, name=%(name)s)'
        return tpl % {
            'class': self.__class__.__name__,
            'function': getattr(self._function, '__name__', repr(self._function)),
            'configuration': repr(self.configuration()),
            'name': repr(self.name),
        }

    def configuration(self):
        return Configuration.merge(*self._configurations)

    def __getitem__(self, section):
        c = self.configuration()
        defaults = {}
        for name, value in chain(c[None].items(), c.get(section, {}).items()):
            defaults[name] = value
        return Function(self._d[section], defaults)

    def __copy__(self):
        return Program(self._function, *self._configurations, name=self.name)

    def __iter__(self):
        for section, function in self._d.items():
            if callable(function):
                yield section

    def __len__(self):
        return len(self._d)

def _sub(section, fs):
    '''
    :param tuple section: Sections that this sub-program is under
    '''
    if hasattr(fs, 'items'):
        items = fs.items()
    elif hasattr(fs, '__iter__'):
        gs = list(fs)
        if all(hasattr(g, '__name__') for g in gs):
            items = ((g.__name__, g) for g in gs)
        else:
            raise TypeError('Bad function component: %s' % repr(fs))
    elif hasattr(fs, '__call__'):
        yield section, fs
        if isinstance(fs, type):
            items = (item for item in getmembers(fs) \
                     if not item[0].startswith('_'))
        else:
            items = []
    else:
        raise TypeError('Bad function component: %s' % repr(fs))

    for name, f in items:
        _validate_param_name(*name)
        yield from _sub(section + (name,), f)

def _validate_param_name(*xs):
    msg = 'It is best if you remove "%s" from the argument name "%s".'
    for x in xs:
        for c in '/?[]<>():':
            if x in c:
                logger.warning(msg % (c, x))



from pprint import pformat

class Loop(object):
    '''
    This class is written strangely so you can wrap a program and thus
    make nice Python documentation for an Interface.
    '''
    def __init__(self, spec):
        self._spec = spec
        update_wrapper(self, self._spec.setup)
        self.__name__ = setup._spec.__name__

    def __call__(self, *args, **kwargs):
        state = self._spec.setup(*args, **kwargs)
        while True:
            try:
                state = self._spec.loop(state)
            except Exit:
                break

def stream(x):
    '''
    :param x: Anything (returned from the underlying program)
    :rtype: Iterable of bytes
    :returns: Output as bytes
    '''
    if x is None:
        raise StopIteration
    elif isinstance(x, bytes):
        yield x
    elif isinstance(x, str):
        yield x.encode('utf-8')
    elif isinstance(x, dict) or not hasattr(x, '__iter__'):
        yield pformat(x).encode('utf-8')
    else:
        for y in x:
            yield from stream(y)



import subprocess
from argparse import RawDescriptionHelpFormatter
from collections import ChainMap, OrderedDict
from triedict import triedict
from . import _bind
from . import _util
from ..annotations import Flag
from .. import exceptions
from ..program import Program

def django(program, handlers=None, freedesktop=False):
    '''
    Produce a django management command to be assigned to
    ``<project>.management.commands.Command``.

    :type program: Program, callable, list, or dict
    :param program: The program for which to produce the interface
    :param dict handlers: Mapping from content type to handler program
    :param bool freedesktop: Fall back to xdg-open if a handler is not available
    '''
    try:
        from django.core.management.base import BaseCommand, CommandError
    except ImportError:
        logger.error('Django is not installed; run this: pip install horetu[all]')
        sys.exit(1)

    if not handlers:
        handlers = {}

    if not isinstance(program, Program):
        program = Program(program)

    if set(program) == {()}:
        function = program[()]
    else:
        raise ValueError('Subcommands are not allowed in the argparse/django interface.')

    parser_hack = []
    class Command(BaseCommand):
        def add_arguments(self, parser):
            parser_hack.append(parser)
            parser.formatter_class=RawDescriptionHelpFormatter
            parser.description = function.description
            one(parser, function)

        def handle(self, *args, **options):
            if args != ():
                raise NotImplementedError('I don\'t know what to do with args.: ' + repr(args))
            inputs = parse_options(function, options)
            fp = self.stdout
            try:
                raw_output = function.run('raw', inputs, {})
                rt = function.return_type
                if rt and (freedesktop or (rt.parsed in handlers)):
                    handler = handlers.get(rt.parsed, XDG_OPEN)
                    if isinstance(raw_output, bytes):
                        stream = [raw_output]
                    else:
                        stream = raw_output
                    p = subprocess.Popen(handler, stdin=subprocess.PIPE, stderr=stderr)
                    for element in stream:
                        if isinstance(element, bytes):
                            p.stdin.write(element)
                        else:
                            raise ValueError('When you set a content type, the function must return either bytes or an iterable of bytes.')
                    p.stdin.close()
                    p.wait()
                else:
                    for line in _util.stream(raw_output):
                        self.stdout.buffer.write(line + b'\n')
                        self.stdout.flush()
        #   except exceptions.CouldNotParse as e:
        #       parser_hack[0].error('')
            except exceptions.Error as e:
                raise CommandError(e.message)
            # It is a bug if other horetu exceptions are raised
    return Command

def parse_options(function, kwargs):
    arguments = OrderedDict()
    params = triedict({param.name: param for param in function.all_parameters()})
    for key, value in kwargs.items():
        if key in params:
            p = params[key]
            if isinstance(p, Flag):
                value = Flag.YES
            if not p.name in arguments:
                arguments[p.name] = []
            arguments[p.name].append(value)
    return arguments

def one(parser, f):
    def kw(param, **kwargs):
        def Type(x):
            return param.loads('raw', x)
        Type.__name__ = param.__name__
        kwargs.update({
            'type': Type,
            'default': param.default,
            'help': param.description,
        })
        if kwargs['action'] in {'store_true', 'store_false', 'count'}:
            del(kwargs['nargs'])
        return kwargs

    for param in f.positional:
        parser.add_argument(param.name, **kw(param, action='store'))
    for param in f.keyword1:
        parser.add_argument(param.name, **kw(param, nargs='?'))
    for param in f.var_positional:
        parser.add_argument(param.name, **kw(param, nargs='*', action='store'))
    for param in f.keyword2:
        if isinstance(param, Flag):
            action = 'store_true'
        elif param.is_list_type:
            action = 'append'
        else:
            action = 'store'
        parser.add_argument('--'+param.name, **kw(param, action=action))



import logging
from itertools import chain
from collections import ChainMap
from .doc import Doc
from ..signature import Signature

logger = logging.getLogger(__name__)

WIDTH = 300
MARGIN = 20
SPINNER = WIDTH - MARGIN * 2

DEFAULT_NAME = 'Horetu program'

def gtk(program,
        handlers=None, freedesktop=False,
        ):
    '''
    Run a GTK application.

    :type program: Program, callable, list, or dict
    :param program: The program for which to produce the interface
    :param dict handlers: Mapping from content type to handler program
    :param bool freedesktop: Fall back to xdg-open if a handler is not available
    '''
    try:
        import gi
        gi.require_version('Gtk', '3.0')
        from gi.repository import Gtk
    except ImportError:
        logger.error('PyGObject is not installed; run this: pip install horetu[gui]')
    else:
        if not isinstance(program, Program):
            program = Program(program)
        if not program.name:
            program.name = DEFAULT_NAME
        class Window(Gtk.Window):
            handlers = handlers or {}
            freedesktop = freedesktop
            program = program
            def __init__(self):
                Gtk.Window.__init__(self, title=program.name)
                main(self, program)
        win = Window(function)
        win.connect("delete-event", Gtk.main_quit)
        win.show_all()
        Gtk.main()



def _init(self, function):
    # Base layout
    self.set_border_width(10)
    box_outer = Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
                        spacing=6)
    self.add(box_outer)

    # Inputs
    listbox = Gtk.ListBox()
    listbox.set_selection_mode(Gtk.SelectionMode.NONE)
    box_outer.pack_start(listbox, True, True, 0)

    s = Signature(function, with_help=False)
    fdoc = Doc(function)
    captions = dict(ChainMap(dict(fdoc.args), dict(fdoc.kwargs)))
    def arg(x):
        listbox.add(_param_row(x, captions[x.name]))
    for x in s.positional:
        arg(x)
    listbox.add(Gtk.HSeparator())
    for x in chain(s.keyword1, s.keyword2.values()):
        arg(x)
#   if s.var_positional:
#       listbox.add(_param_row(s.var_positional))

    # Output
    stack = Gtk.Stack()
    
    spinner = Gtk.Spinner()
    spinner.set_size_request(SPINNER, SPINNER)
    spinner.start()

    button_go = Gtk.ToolButton()
    button_go.set_icon_name('media-playback-start')
    button_go.connect('clicked', lambda x:
        stack.set_visible_child(spinner))

    scrolledwindow = Gtk.ScrolledWindow()
    scrolledwindow.set_hexpand(True)
    scrolledwindow.set_vexpand(True)
    self.textview = Gtk.TextView()
    self.textbuffer = self.textview.get_buffer()
    self.textbuffer.set_text('These are the output messages')
    scrolledwindow.add(self.textview)

    stack.add_titled(button_go, "go", "Go Button")
    stack.add_titled(spinner, "running", "Running")
    stack.add_titled(scrolledwindow, "output", "Program Output")
    box_outer.pack_start(stack, True, True, 0)

def _param_row(param, documentation):
    vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
    label1 = Gtk.Label(xalign=0)
    label1.set_markup("<b>%s</b>" % param.name)
    label2 = Gtk.Label(xalign=0)
    label2.set_markup("<i>%s</i>" % documentation)
    vbox.pack_start(label1, True, True, 0)
    vbox.pack_start(label2, True, True, 0)

    row = Gtk.ListBoxRow()
    hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=50)
    row.add(hbox)
    hbox.pack_start(vbox, True, True, 0)

    if param.takes_parameter:
        pass
    else:
        switch = Gtk.Switch()
        switch.props.valign = Gtk.Align.CENTER
        hbox.pack_start(switch, False, True, 0)

    return row



from .program import Program
from .exceptions import Error, ShowHelp
from .config import Configuration
from .interfaces.cli import cli
from .interfaces.irc import irc
from .interfaces.wsgi import (
    plain as wsgi_plain,
    form as wsgi_form,
)
from .interfaces.django import django
from .interfaces.config import (
    default as config_default,
    run as config_run,
)



import os
import abc

class BaseHoretuException(Exception, metaclass=abc.ABCMeta):
    ok = True
    def __init__(self, message=None, section=tuple(), param=None):
        # message first so it is easy not to provide a section
        if section:
            self.section = section
        else:
            self.section = tuple()
        self.message = message
        self.param = param

    def __repr__(self):
        return '%s(%r, section=%r, param=%r)' % (
            self.__class__.__name__,
            self.message,
            self.section,
            self.param,
        )
    __str__ = __repr__

class ShowHelp(BaseHoretuException):
    pass

class Config(BaseHoretuException):
    def __init__(self, filename):
        self.filename = os.path.expanduser(filename)

class Exit(BaseHoretuException):
    pass

class BaseHoretuError(BaseHoretuException, metaclass=abc.ABCMeta):
    ok = False

class AnnotationNotImplemented(BaseHoretuException):
    pass

class CouldNotParse(BaseHoretuError):
    pass

class Error(BaseHoretuError):
    pass



from xml.etree import ElementTree as E

SUBMIT = E.fromstring(b'<button type="submit">Run</button>')

def form(function, method, arguments, message, output):
    '''
    :param dict Configuration flatconf: Flattened configuration,
        with function call input values
    :param Function func: Function for the given endpoint
    :param bool with_help: Include help text
    :param str output: Program output
    '''
    if not isinstance(output, str):
        raise TypeError('Form output must be str')
    m = E.Element('textarea', **{'readonly': '', 'class': 'message list-box-row'})
    if message:
        m.text = str(message)

    i = E.Element('form', method=method, enctype='multipart/form-data', action='')
    for param in function.all_parameters():
        try:
            element = param.dumps('wsgi', param.default)
        except Exception as e:
            logger.exception('Failed to dump value for %s' % param.name)
        else:
            i.append(element)
    i.append(SUBMIT)

    o = E.Element('textarea', **{'readonly': '', 'class': 'output list-box-row'})
    if output:
        o.text = output

    main = E.Element('main', role='main')
    main.append(m)
    main.append(i)
    main.append(o)
    return E.tostring(main, method='html').decode('utf-8')

default_template = '''\
<!DOCTYPE html>
<html>
  <head>
    <title>${title}</title>
    <meta content="${description}" name="description">
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
    <meta charset="utf-8">
    <meta content="horetu" name="generator">
    <meta content="${author}" name="author">
    <style>
      .list-box-row {
        width: 640px;
        display: block;
      }
      .list-box-row > * { padding: 0.25em; margin: 0.25em; }

      .message { height: 2em; color: red; }
      .message, .output { background-color: #eee; }
      .output { margin-top: 1em; height: 6em; }

      /* Single-row field */
      .list-box-row.oneline label { float: left; }
      .list-box-row.oneline { height: 4em; }
      .list-box-row.oneline input, .list-box-row.oneline select
        { height: 3em; float: right; }

      /* Multi-row field */
      .list-box-row.multiline { height: 8em; }
      .list-box-row textarea { display: block; height: 6em; }
    </style>
  </head>
  <body>${body}</body>
</html>
'''



import os
import logging
import shlex
from io import BytesIO
from traceback import format_exc
from .cli import cli as _horetu_cli
from ._bind import web_section
from ._util import stream
from .. import exceptions
from ..program import Program
from ..annotations import Port
try:
    from irc.strings import lower
except ImportError:
    pass

logger = logging.getLogger(__name__)

def _irc(program, server, channel, nick, port, start=True, IRCBot=None):
    '''
    :param bool start: Whether to start the bot. The main reason not to is
        for testing.
    :param type IRCBot: IRCBot to subclass. The default is
        :py:class:`irc.bot.SingleServerIRCBot`. The main reason to change it is
        for testing.
    '''
    from irc.client import ip_numstr_to_quad, ip_quad_to_numstr
    if not IRCBot:
        from irc.bot import SingleServerIRCBot as IRCBot

    if not isinstance(program, Program):
        program = Program(program)
    if not nick:
        nick = program.name
    if not (server and channel and nick):
        raise ValueError('You must set the server, channel, and nick.')

    class HoretuBot(IRCBot):
        def on_nicknameinuse(self, c, e):
            c.nick(c.get_nickname() + "_")

        def on_welcome(self, c, e):
            c.join(channel)

        def on_privmsg(self, c, e):
            _respond(program, e.source.nick, c.notice, c.get_nickname(), e.arguments[0])

        def on_pubmsg(self, c, e):
            line = e.arguments[0]
            nickname = c.get_nickname()
            zero = line.lstrip().split()[0]
            mentioned = zero.startswith(nickname)
            if mentioned:
                _respond(program, e.target, c.privmsg, zero, line)

    bot = HoretuBot([(server, port)], nick, nick)
    if start:
        bot.start()
    else:
        return bot

def irc(program, server, channel,
        nick=None, *, port: Port=6667,
        cli=False, only_mentions=False, debug=False):
    '''
    Render a Python function as an IRC bot.

    :type program: Program, callable, list, or dict
    :param program: The program for which to produce the interface
    :param server: IRC server to connect to
    :param channel: IRC channel to join
    :param nick: IRC handle
    :param int port: IRC server port
    :param bool cli: If True, render a command-line program that starts an IRC
        bot with the configured values as defaults.
    '''
    if cli:
        def wrapper(server=server, channel=channel,
                    nick=nick, *, port: Port=port):
            _irc(program, server, channel, nick, port)
        wrapper.__doc__ = '''
        Run %s as an IRC bot.

        :param server: IRC server to run on
        :param channel: Channel on the IRC server to run on
        :param nick: Nick to connect with
        :param int port: Port on the server to connect to
        ''' % (program.name or 'the program')

        name = program.name if isinstance(program, Program) else None
        _horetu_cli(Program(wrapper, name))
    else:
        _irc(program, server, channel, nick, port)

def _respond(program, target, method, nickname, line):
    def success(x):
        method(target, x.decode('utf-8'))
    def error(x):
        method(target, x.decode('utf-8'))

    class Exit(Exception):
        pass
    def _exit(code=0):
        raise Exit

    out = BytesIO()
    try:
        _horetu_cli(program, argv=[nickname] + shlex.split(line),
                    exit_=_exit, stdout=out, stderr=out)
        success(out.getvalue().decode('utf-8'))
    except Exit:
        pass
    except UnicodeDecodeError:
        error(b'error: Bad bot output encoding')



import logging
import inspect
import re
from collections import OrderedDict
from itertools import filterfalse, chain
from functools import partial
from enum import Enum
from .annotations import Help, Annotation
from .exceptions import CouldNotParse
from ._content_type import ContentType

logger = logging.getLogger(__name__)

class Function(object):
    def __init__(self, f, configured_defaults):
        '''
        :param dict configured_defaults: Defaults from program configuration
        '''
        self._function = f
        self._configured_defaults = configured_defaults or {}

        s = inspect.signature(f)
        for kind in Kind:
            setattr(self, kind.name, [])
        
        for kind, param in _kinds(s):
            getattr(self, kind.name).append(Annotation.bind(param))

        for param in self.all_parameters():
            _inspect_param = s.parameters[param.name]
            if _inspect_param.default != _inspect_param.empty:
                param.default = _inspect_param.default

        ps = {param.name: param for param in self.all_parameters()}
        doc_params, self.description = _doc(f)
        for name, desc in doc_params:
            if name in ps:
                ps[name].description = desc
            else:
                logger.warning('Docstring references non-variable "%s"' % name)

        if s.return_annotation in {s.empty, None}:
            self.return_type = None
        else:
            self.return_type = ContentType(s.return_annotation)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self._function)

    def all_positionals(self):
        return chain(*(getattr(self, kind.name) for kind in
                       [Kind.positional, Kind.keyword1, Kind.var_positional]))
    def all_parameters(self):
        return chain(*(getattr(self, kind.name) for kind in Kind))

    def help_flag(self):
        for param in self.keyword2:
            if isinstance(param, Help):
                return param.name

    def _manual_help_flag(self):
        s = inspect.signature(self._function)
        for param in s.parameters.values():
            if param.annotation == Help:
                return param.name

    def require_help(self):
        s = inspect.signature(self._function)
        if (Help.NAME not in s.parameters) and not self._manual_help_flag():
            class param:
                name = Help.NAME
                annotation = Help
                default = False
                empty = inspect.Parameter.empty
                kind = 'beep'
                VAR_POSITIONAL = 'boop'
            self.keyword2.append(Annotation.bind(param))

    @property
    def defaults(self):
        '''
        Assign defaults from the function signature.
        No parsing nor reshaping is needed.
        '''
        arguments = OrderedDict()
        for param in self.all_parameters():
            if param.default != inspect.Parameter.empty:
                arguments[param.name] = param.default 
        return arguments

    def run(self, loader, raw_arguments, var_keyword):
        '''
        :param loader: "raw", "wsgi", or "gtk"
        :type raw_arguments: dict of str to list of unparsed values
        :param raw_arguments: Arguments of function call
        '''
        arguments = {}

        # Parse inputs, convert list-typed status, and assign
        for param in self.all_parameters():
            def _parse(arg):
                f = partial(param.loads, loader)
                try:
                    if param.is_list_type:
                        return param.finalize_list(list(map(f, arg)))
                    else:
                        return f(arg[-1])
                except ValueError as e:
                    raise CouldNotParse('Could not parse %s: %s' % (param.name, e))

            if isinstance(param, Help):
                if raw_arguments.get(param.name,
                        self._configured_defaults.get(param.name,
                            self.defaults[param.name])):
                    param.loads(loader, True)
            elif param.name in raw_arguments:
                arguments[param.name] = _parse(raw_arguments[param.name])
            elif param.name in self._configured_defaults:
                arguments[param.name] = _parse(self._configured_defaults[param.name])
            elif param.name in self.defaults:
                arguments[param.name] = self.defaults[param.name]

        # Checks
        for param in self.positional:
            if param.name not in arguments:
                raise CouldNotParse('Missing positional argument: %s' % param.name)

        # Represent as args and kwargs
        args = []
        kwargs = {}

        for param in self.positional + self.keyword1 + self.var_positional:
            include = {True: args.extend, False: args.append}[param.is_list_type]
            if param.name in arguments:
                include(arguments[param.name])

        for param in self.keyword2:
            if param.name in arguments:
                if param.is_list_type:
                    kwargs[param.name] = list(arguments[param.name])
                else:
                    kwargs[param.name] = arguments[param.name]

        if var_keyword:
            param = self.var_keyword[0]
            for key, value in var_keyword.items():
                if key in arguments:
                    raise CouldNotParse('This is already an argument: %s' % key)
                else:
                    kwargs[key] = param.loads(loader, value)

        # Run
        return self._function(*args, **kwargs)


COLON = re.compile(r'^[\s:]')
def _colon(line):
    return bool(re.match(COLON, line))

def _desc(lines):
    return '\n'.join(filterfalse(_colon, lines)).strip()

PARAM = re.compile(r'^:param (?:[^:]+ )?([^:]+): (.+)$')
INDENT = re.compile(r'\s+([^\s].*)')

def _doc(function):
    d = inspect.getdoc(function)
    if d:
        lines = d.split('\n')
    else:
        lines = []

    y = []
    z = ''

    name = None
    desc = ''
    prevline_empty = None
    for line in lines:
        if _colon(line):
            m = re.match(PARAM, line)
            n = re.match(INDENT, line)
            if m:
                # New parameter
                if name:
                    # Save the previous parameter if it exists.
                    y.append((name, desc.strip()))
                    name = None

                # Start the new description
                name, desc = m.groups()
            elif name and n:
                desc += ' ' + n.group(1)
        elif line.strip():
            if prevline_empty == None:
                pass
            elif prevline_empty == True:
                z += '\n\n'
            elif prevline_empty == False:
                z += ' '
            z += line.strip()
            prevline_empty = False
        else:
            prevline_empty = True

    if name:
        # Save the last parameter
        y.append((name, desc.strip()))
    
    return y, z

class Kind(Enum):
    positional = 1
    keyword1 = 2
    var_positional = 3
    keyword2 = 4
    var_keyword = 5

    # https://docs.python.org/3/library/enum.html#orderedenum
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        else:
            return NotImplemented
    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        else:
            return NotImplemented
    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        else:
            return NotImplemented
    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        else:
            return NotImplemented

KINDS = {
    Kind.positional: {
        inspect.Parameter.POSITIONAL_ONLY,
        inspect.Parameter.POSITIONAL_OR_KEYWORD
    },
    Kind.keyword1: {inspect.Parameter.POSITIONAL_OR_KEYWORD},
    Kind.var_positional: {inspect.Parameter.VAR_POSITIONAL},
    Kind.keyword2: {inspect.Parameter.KEYWORD_ONLY},
    Kind.var_keyword: {inspect.Parameter.VAR_KEYWORD},
}

def _kinds(sig):
    has_k2 = Kind.keyword2 in (k for k,_ in _naive_kinds(sig))
    for kind, param in _naive_kinds(sig):
        if kind == Kind.keyword1 and not has_k2:
            yield Kind.keyword2, param
        else:
            yield kind, param

def _naive_kinds(sig):
    for param in sig.parameters.values():
        if param.kind in KINDS[Kind.positional].union(KINDS[Kind.keyword1]):
            if param.default == param.empty:
                kind = Kind.positional
            else:
                kind = Kind.keyword1
        elif param.kind in KINDS[Kind.var_positional]:
            kind = Kind.var_positional
        elif param.kind in KINDS[Kind.keyword2]:
            kind = Kind.keyword2
        elif param.kind in KINDS[Kind.var_keyword]:
            kind = Kind.var_keyword
        yield kind, param



'''
Everything is string type, even if this is inefficient.

Class.loads always takes str input, and Class.dumps always
returns str output.
'''
import logging
import os
import io
import re
from functools import partial
from collections import OrderedDict
from xml.etree import ElementTree as E
from abc import ABCMeta
from . import exceptions

# Just for checking types
import decimal, inspect

logger = logging.getLogger(__name__)

UNDOCUMENTED = 'Undocumented'

_default_true = 'bool arguments must have default of False; consider changing param %(name)s to no%(name)s'

class AnnotationFactory(object):
    def __init__(self, Class, *args, **kwargs):
        self._Class = Class
        self._args = args
        self._kwargs = kwargs
    def __call__(self):
        try:
            return self._Class(*self._args, **self._kwargs)
        except ValueError:
            raise exceptions.CouldNotParse

class Annotation(object, metaclass=ABCMeta):
    allow_get = True
    def __init__(self, *args, **kwargs):
        raise NotImplementedError
    @staticmethod
    def finalize_list(xs):
        return xs

    def loads(self, loader, data):
        return {
            'raw': self._loads_raw,
            'wsgi': self._loads_wsgi,
            'urwid': self._loads_urwid,
        }[loader](data)
    def dumps(self, dumper, value):
        return {
            'raw': self._dumps_raw,
            'wsgi': self._dumps_wsgi,
            'urwid': self._dumps_urwid,
        }[dumper](value)

    def _label(self, tree, lines='one'):
        parent = E.Element('div', **{'class': 'list-box-row %sline' % lines})

        label = E.Element('label', **{
            'for': self.name,
            'title': self.description,
        })
        label.text = self.name

        parent.append(label)
        parent.append(tree)
        return parent

    #: Default documentation string
    description = UNDOCUMENTED
    default = inspect.Parameter.empty

    @classmethod
    def factory(Class, *args, **kwargs):
        return AnnotationFactory(Class, *args, **kwargs)

    @classmethod
    def bind(Class, param, _annotation=None):
        '''
        Bind an annotation to a parameter.

        :param inspect.Parameter param: Parameter from the function signature
        :param _annotation: Annotation for the resulting object instead
            of param.annotation, used only internally for recursion
        :ivar bool is_list_type: Whether the parameter takes multiple arguments
        :ivar default: Default value for keyword arguments (used for
            generating default configuration files)
        '''
        if _annotation:
            a = _annotation
        else:
            a = param.annotation

        if      (
                    (a == bool) or \
                    (a == param.empty and 
                        isinstance(param.default, bool) and
                        param.default==True)
                ) and \
                (not isinstance(param.annotation, list)) and \
                (param.default != False):
            raise ValueError(_default_true % {'name': param.name})
        elif isinstance(a, Class):
            raise TypeError('Annotations must be Annotation subclasses, not instances.')
        elif isinstance(a, type) and issubclass(a, Annotation):
            obj = a() # This works, of course, for only Annotations that accept no arguments.
        elif isinstance(a, AnnotationFactory):
            obj = a()
        elif a == inspect.Parameter.empty and \
                isinstance(param.default, bool) and param.default == False:
            obj = Boolean()
        elif isinstance(a, list) and len(a) == 1:
            if isinstance(a[0], list) and len(a[0]) == 1:
                raise ValueError('You can\'t nest list annotations.')
            obj = Class.bind(param, _annotation=a[0])
        elif _is_hashable(a) and a in annotations_map:
            obj = annotations_map[a]()
        elif hasattr(a, 'loads') and hasattr(a, 'dumps') \
            or hasattr(a, 'load') and hasattr(a, 'dump'):
            obj = Encoder(a)
        elif hasattr(a, '__call__'):
            obj = String(a)
        elif hasattr(a, 'items'):
            obj = FactorMapping(a)
        elif isinstance(a, tuple):
            obj = Factor(a)
        else:
            raise ValueError('Bad input annotation: %s' % a)

        if hasattr(obj, '__name__'):
            pass
        elif hasattr(a, '__name__'):
            obj.__name__ = a.__name__
        else:
            obj.__name__ = obj.__class__.__name__

        obj.name = param.name
        obj.default = param.default
        obj.is_list_type = (param.kind == param.VAR_POSITIONAL) or \
            (isinstance(a, list) and len(a) == 1)

        return obj

    def _loads_raw(self, data):
        raise NotImplementedError
    def _dumps_raw(self, value):
        raise NotImplementedError

    def _loads_wsgi(self, data):
        return self._loads_raw(data)
    def _dumps_wsgi(self, value):
        raise NotImplementedError
    
    def _loads_urwid(self, data):
        return self._loads_raw(data)
    def _dumps_urwid(self, value):
        raise NotImplementedError

    def __str__(self):
        return '<%s: %s>' % (self.name, self.__name__)

    YES = 'yes'
    NO = 'no'

def _is_hashable(x):
    '''
    I implemented this because ``isinstance({}.__getitem__, collections.Hashable)``
    does not do what I want.
    '''
    try:
        hash(x)
    except TypeError:
        return False
    else:
        return True

def _d(f):
    def wrapper(x):
        if x != inspect.Parameter.empty:
            try:
                return f('raw', x)
            except Exception:
                logger.exception('Error in dumping to web form')
        return ''
    return wrapper

class Encoder(Annotation, metaclass=ABCMeta):
    def __init__(self, x):
        if hasattr(x, 'loads') and hasattr(x, 'dumps'):
            self._loads_raw = x.loads
            self._dumps_raw = x.dumps
        elif hasattr(x, 'load') and hasattr(x, 'dump'):
            t = getattr(self, '_type', getattr(x, '_type'))
            IO = {str: io.StringIO, bytes: io.BytesIO}[t]

            def _loads_raw(text):
                with IO(text) as fp:
                    obj = x.load(fp)
                return obj
            self._loads_raw = _loads_raw

            def _dumps_raw(obj):
                with IO() as fp:
                    x.dump(obj, fp)
                    text = fp.getvalue()
                return text
            self._dumps_raw = _dumps_raw
        else:
            raise TypeError('Not an encoder: %s' % x)
        self._x = x # just for repr
    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self._x)

    def _dumps_wsgi(self, value):
        '''
        :param param: The parameter
        :param value: The default value
        '''
        tree = E.Element('textarea', name=self.name)
        if value:
            tree.text = _d(self.dumps)(value)
        return self._label(tree, lines='multi')

# Flags
class Flag(Annotation):
    def __init__(self, loads):
        self._loads_raw = loads
    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self._loads_raw)

    def _loads_raw(self, yes=True):
        if yes in {self.YES, True}:
            self._loads()
            return True
        else:
            return False
    def _dumps_raw(self, yes):
        return self.YES if yes else self.NO

    def _dumps_wsgi(self, yes):
        attr = {'type': 'checkbox', 'name': self.name}
        if yes:
            attr['checked'] = self.YES
        input_ = E.Element('input', **attr)
        return self._label(input_)

@Flag.factory
def Boolean(_=None):
    return True
class Count(Flag):
    def __init__(self):
        super(Count, self).__init__(lambda _=None: True)
    finalize_list = len

class Help(Flag):
    NAME = 'help'
    __name__ = 'Help'
    def __init__(self):
        pass

    def __repr__(self):
        return '%s()' % self.__class__.__name__

    description = 'Display this help.'
    default = False

    def _loads_raw(self, _):
        raise exceptions.ShowHelp

# String types
class Text(Annotation, metaclass=ABCMeta):
    def _dumps_wsgi(self, value):
        attr = {'type': 'text', 'name': self.name}
        if value:
            attr['value'] = _d(self.dumps)(value)
        input_ = E.Element('input', **attr)
        return self._label(input_)

class Identity(Text):
    def __init__(self, type=str):
        self._type = type
    def __repr__(self):
        return '%s(type=%s)' % (self.__class__.__name__, self._type)

    def _check_type(self, x):
        if self._type and not isinstance(x, self._type):
            p = (self._type, type(x), repr(x))
            raise TypeError('Value must be %s, not %s: %s' % p)
        else:
            self._assign_type(x)
    def _assign_type(self, x):
        if not self._type:
            self._type = type(x)

    def _loads_raw(self, data):
        self._check_type(data)
        return data
    def _dumps_raw(self, value):
        self._check_type(value)
        return value

class Config(Identity):
    def __init__(self, filename):
        self._default_filename = filename
        super(Config, self).__init__(type=str)
    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.default_filename)

    def _loads_raw(self, _):
        raise exceptions.Config(self._default_filename)
    def _loads_wsgi(self, data):
        raise ValueError('Do not use the Config annotation for websites.')
    def _dumps_wsgi(self, obj):
        raise ValueError('Do not use the Config annotation for websites.')

class String(Text):
    def __init__(self, loads=lambda x: x):
        if isinstance(loads, partial):
            self.__name__ = loads.func.__name__
        else:
            self.__name__ = loads.__name__
        self._loads_raw = loads
        self._dumps_raw = str
    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self._loads_raw.__name__)

class Regex(Text):
    def __init__(self, expr):
        def loads(x):
            if re.match(expr, x):
                return x
            else:
                raise ValueError('Does not match regular expression: %s' % expr)

        self._expr = re.compile(expr)
        self.__name__ = '<Regex %s>' % self._expr
        self._loads_raw = loads
        self._dumps_raw = str
    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self._expr)

class Bytes(Text):
    def __init__(self, loads=lambda x: x, dumps=lambda x: x, encoding='ascii'):
        decode, encode = _coders(encoding)

        self._loads_raw = lambda x: loads(encode(x))
        self._dumps_raw = lambda x: decode(dumps(x))

        # Just for repr
        self._loads = loads
        self._dumps = dumps
        self._encoding = encoding
    def __repr__(self):
        return '%s(loads=%s, dumps=%s, encoding=%r)' % \
            (self.__class__.__name__, self._loads, self._dumps, self._encoding)

def _coders(encoding):
    if isinstance(encoding, str):
        loads = lambda x: bytes.decode(x, encoding)
        dumps = lambda x: str.encode(x, encoding)
    elif encoding == None:
        loads = bytes.decode
        dumps = str.encode
    else:
        raise TypeError('Must be None or str, not %s' % type(encoding))
    return loads, dumps

@String.factory
def InputDirectory(x):
    if os.path.isdir(x):
        return x
    else:
        raise ValueError('No such directory: ' + str(x))

@String.factory
def OutputDirectory(x):
    os.makedirs(x, exist_ok=True)
    return x

@String.factory
def Port(x):
    y = int(x)
    if 1 <= y <= 65535:
        return y
    else:
        raise ValueError('Not a valid port number: %d' % y)

Decimal = String.factory(decimal.Decimal)
Integer = String.factory(int)
Float = String.factory(float)

class Range(String):
    def __init__(self, minimum, maximum, loads=float, dumps=str, step='any'):
        self._loads_raw = loads
        self._dumps_raw = dumps
        self._min = minimum
        self._max = maximum
        self._step = step
    def __repr__(self):
        return '%s(%r, %r, loads=%s, dumps=%s, step=%r)' % \
            (
                self.__class__.name, self._min, self._max,
                self._loads_raw, self._dumps_raw, self._step,
            )
        
    def _dumps_wsgi(self, name, default):
        '''
        :param name: The parameter name
        :param default: The default value, ignored
        '''
        attr = {
            'type': 'range', 'name': name,
            'min': _d(self.dumps)(self._min), 'max': _d(self.dumps)(self._min),
            'step': self._step if isinstance(self._step, str) else _d(self.dumps)(self._step),
        }
        if default:
            attr['value'] = _d(self.dumps)(default)
        input_ = E.Element('input', **attr)
        return self._label(input_)

# Factor types
class FactorMapping(String):
    def __init__(self, options):
        if not hasattr(options, 'items'):
            raise ValueError('Options must be dict-like')
        if not all(isinstance(option, str) for option in options):
            raise ValueError('All options must be of str type.')
        self._options = options

    def _reverse_options(self, value):
        xs = list(k for (k,v) in self._options.items() if v == value)
        if 0 == len(xs):
            raise ValueError('Must be one of %s' % repr(tuple(xs)))
        else:
            if 1 < len(xs):
                logger.warning('Multiple keys for %s, using the first' % repr(value))
            return xs[0]

    def _loads_raw(self, key):
        if key in self._options:
            return self._options[key]
        else:
            raise ValueError('Must be one of %s' % repr(tuple(self._options)))

    def _dumps_raw(self, value):
        return self._reverse_options(value)

    def _dumps_wsgi(self, selected):
        select = E.Element('select', name=self.name)
        if self.is_list_type:
            select.set('multiple', self.YES)
        else:
            selected = {selected}

        for value, text in self._options.items():
            option = E.Element('option', value=value)
            option.text = text
            if value in selected:
                option.set('selected', self.YES)
            select.append(option)

        return self._label(select)

class Factor(FactorMapping):
    def __init__(self, options):
        super(Factor, self).__init__(OrderedDict([(o,o) for o in options]))

# File pointer types
class File(Annotation):
    allow_get = False
    def __init__(self, mode, encoding=None, **kwargs):
        self._mode = mode
        self._decode, self._encode = _coders(encoding)
        self._encoding = encoding
        self._kwargs=kwargs

    def _loads_raw(self, filename):
        if 'r' in self._mode and not os.path.isfile(filename):
            raise ValueError('No such file: %s' % filename)
        return open(filename, mode=self._mode, encoding=self._encoding, **self._kwargs)

    def _dumps_raw(self):
        return self.fp.name

    def _loads_wsgi(self, data):
        if 'b' in self._mode:
            return data.file
        else:
            return io.StringIO(self._decode(data.file.read()))

    def _dumps_wsgi(self, default):
        '''
        :param name: The parameter name
        :param default: The default value, ignored
        '''
        attr = {'type': 'file', 'name': self.name, 'size': '40'}
        input_ = E.Element('input', **attr)
        return self._label(input_)

InputFile = File.factory('r')
InputBinaryFile = File.factory('rb')
OutputFile = File.factory('w')
OutputBinaryFile = File.factory('wb')

annotations_map = {
    bool: Boolean,
    bytes: Bytes,
    decimal.Decimal: Decimal, 'Decimal': Decimal,
    float: Float,
    inspect.Parameter.empty: Identity,
    int: Integer,
    str: Identity,

    'Help': Help,
    'Count': Count,

    'Config': Config,
    'Port': Port,
    'Range': Range,

    'InputFile': InputFile,
    'InputBinaryFile': InputBinaryFile,
    'OutputFile': OutputFile,
    'OutputBinaryFile': OutputBinaryFile,
}



import sys
import logging
from collections import ChainMap
from functools import partial
from . import _util
from ..program import Program
from ..annotations import OutputFile
from ..exceptions import AnnotationNotImplemented

logger = logging.getLogger(__name__)

def default(program, with_sections=False):
    '''
    Write a configuration file with a program's defaults.

    :type program: Program, callable, list, or dict
    :param program: The program for which to produce the interface,
    :param bool with_sections: Give each subcommand its own namespace rather
        than putting all of the arguments in a global namespace.
    '''
    if not isinstance(program, Program):
        program = Program(program)
    c = program.configuration()

    for section in program:
        function = program[section]
        params = {param.name: param for param in function.all_parameters()}
        for key, x in function.defaults.items():
            if not key in c:
                param = params[key]
                f = partial(param.dumps, 'raw')
                try:
                    if param.is_list_type:
                        c[section].extend(key, map(f, x))
                    else:
                        c[section].append(key, f(x))
                except AnnotationNotImplemented:
                    logger.warning('No dumps method for %s annotation, skipping' % key)

    return c.write(with_sections=with_sections)

def run(program, section=tuple()):
    '''
    Run a program, applying arguments from any included configuration files.

    :type program: Program, callable, list, or dict
    :param program: The program for which to produce the interface,
    :type section: Iterable of tuple of str
    :param section: Subcommand to run.
    '''
    if not isinstance(program, Program):
        program = Program(program)
    return program[section].run('raw', [], {})



import logging
from traceback import format_exc
from functools import wraps
from io import StringIO
from cgi import FieldStorage
from collections import OrderedDict
from triedict import triedict, KeyNotFound
from string import Template
from .template_plain import man, usage, ensure_bytes
from . import template_form
from .. import _bind
from .._util import stream
from ... import exceptions
from ...annotations import Flag, InputFile
from ...program import Program

logger = logging.getLogger(__name__)

def plain(program, debug=False):
    positionals_in_path = True # TODO Add a False option.

    if not isinstance(program, Program):
        program = Program(program)

    @_as_app(program, debug)
    def app(req, res):
        section, tail = _get_section_tail(program, req)
        function = program[section]
        function.require_help()
        if positionals_in_path:
            try:
                arguments = _args_kwargs(function, tail, _params(function, req))
            except exceptions.CouldNotParse as e:
                e.section = section
                raise e
        else:
            arguments = _no_tail(function, section, tail, req)

        if function.return_type:
            content_type = function.return_type.original
        else:
            logger.warning('%r does not have a return type, using application/octet-stream' % function)
            content_type = 'application/octet-stream'

        try:
            output = function.run('wsgi', arguments, {})
        except exceptions.ShowHelp as e:
            res.content_type = 'text/html; charset=UTF-8'
            res.app_iter = [man(program, e)]
        else:
            res.content_type = content_type
            if function.return_type and function.return_type.parsed.startswith('text/'):
                res.app_iter = stream(output)
            else:
                if isinstance(output, bytes):
                    res.app_iter = [output]
                else:
                    res.app_iter = ensure_bytes(output)
    return app

def form(program, debug=False,
         template: InputFile=StringIO(template_form.default_template),
         description=None, author=None):
    '''
    :type f: Program, callable, list, or dict
    :param f: The program to produce the interface
    :param str name: Name of the program, used for environment variable settings
        If it is ``None`` (the default), we attempt to get the name from the function.
    :param bool debug: Print errors and debugging information to web pages?
    :type template: File-like object
    :param template: Template file for form pages
    :rtype: function
    :returns: WSGI application
    '''
    if not isinstance(program, Program):
        program = Program(program)

    if description is None and set(program) == {()}:
        description = program[()].description
    tpl = Template(template.read())

    @_as_app(program, debug)
    def app(req, res):
        section, tail = _get_section_tail(program, req)
        function = program[section]
        arguments = _no_tail(function, section, tail, req)

        method = 'get' if _allow_get(function) else 'post'
        def render_form(message, output):
            main = template_form.form(function, method, arguments, message or '', output or '')
            res.content_type = 'text/html; charset=UTF-8'
            res.app_iter = [tpl.substitute(
                title=program.name or '',
                description=description or '',
                author=author or '',
                body=main,
            ).encode('utf-8')]

        if req.method == 'GET' and not req.GET:
            render_form(None, None)
        else:
            try:
                output = function.run('wsgi', arguments, {})
            except exceptions.CouldNotParse as e:
                res.status = 400
                output = None
                message = e.message
            except Exception:
                if debug:
                    res.status = 500
                    message = 'Internal server error'
                    output = format_exc()
                else:
                    raise
            else:
                message = None

            if function.return_type:
                res.content_type = function.return_type.original
                res.app_iter = ensure_bytes(output)
            else:
                render_form(message, output)
    return app

def _get_section_tail(program, req):
    if req.path_info == '/':
        section = tuple()
        tail = []
    else:
        section, tail = _bind.web_section(program, req.path_info.split('/')[1:])

    if not section in program:
        raise exceptions.CouldNotParse(section=section, message='No such page')
    return section, tail

def _allow_get(function):
    return all(param.allow_get for param in function.all_parameters())
def _params(function, req):
    return req.params if _allow_get(function) else req.POST

def _no_tail(function, section, tail, req):
    try:
        if 0 < len(tail):
            raise exceptions.CouldNotParse(message='Page not found')
        arguments = _just_kwargs(function, _params(function, req))
    except exceptions.CouldNotParse as e:
        e.section = section
        raise e
    return arguments

def _as_app(program, debug):
    try:
        from webob import Response, Request
    except ImportError:
        logger.error('WebOb is not installed; run this: pip install horetu[all]')
        import sys
        sys.exit(1)

    def decorator(inner_app):
        def app(environ, start_response):
            req = Request(environ)
            res = Response()
            logger.debug(req)

            try:
                res.status = 200
                inner_app(req, res)
            except exceptions.CouldNotParse as e:
                res.status = 404
                res.content_type = 'text/html; charset=UTF-8'
                res.app_iter = [usage(program, e)]
            except Exception:
                if debug:
                    res.status = 500
                    res.content_type = 'text/plain; charset=UTF-8'
                    res.app_iter = [format_exc().encode('utf-8')]
                else:
                    raise
            return res(environ, start_response)
        return app
    return decorator

def _args_kwargs(function, args, kwargs):
    arguments = _bind.positionals(function, args)
        
    keyword2_params = triedict({param.name: param for param in function.keyword2})
    for key, value in kwargs.items():
        try:
            p = keyword2_params[key]
        except KeyNotFound:
            raise exceptions.CouldNotParse('No such argument: %s' % key)
        else:
            if isinstance(p, Flag):
                value = Flag.YES
            if not p.name in arguments:
                arguments[p.name] = []
            arguments[p.name].append(value)
    return arguments

def _just_kwargs(function, post):
    arguments = OrderedDict()
    params = triedict({param.name: param for param in function.all_parameters()})
    for key, value in post.items():
        try:
            p = params[key]
        except KeyNotFound:
            raise exceptions.CouldNotParse('No such argument: %s' % key)
        else:
            if isinstance(p, Flag):
                value = Flag.YES
            if isinstance(value, FieldStorage) or value:
                if not p.name in arguments:
                    arguments[p.name] = []
                arguments[p.name].append(value)
    return arguments



import os
import csv
import logging
from io import StringIO
from collections import OrderedDict

logger = logging.getLogger(__name__)

class Colon(csv.Dialect):
    delimiter = ':'
    doublequote = False
    escapechar = '\\'
    lineterminator = '\n'
    quotechar = None
    quoting = csv.QUOTE_NONE
    skipinitialspace = True
    strict = False

    @classmethod
    def reader(Class, fp):
        return csv.reader(fp, dialect=Class)

    @classmethod
    def writer(Class, fp):
        return csv.writer(fp, dialect=Class)

class Space(Colon):
    delimiter = ' '

    @classmethod
    def read(Class, strrow):
        x = strrow.rstrip(Class.lineterminator) + Class.lineterminator
        with StringIO(x) as fp:
            section = tuple(next(csv.reader(fp, dialect=Class)))
        return section

    @classmethod
    def write(Class, row):
        with StringIO() as fp:
            w = csv.writer(fp, dialect=Class)
            w.writerow(row)
            strrow = fp.getvalue().rstrip(Class.lineterminator)
        return strrow

class Configuration(OrderedDict):
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            first_arg = args[0]
        else:
            first_arg = None

        if isinstance(first_arg, Configuration):
            self.update(first_arg)
        elif isinstance(first_arg, str):
            if os.path.isfile(first_arg):
                with open(first_arg) as fp:
                    self.update(Configuration.read_file(fp))
        elif hasattr(first_arg, 'read'):
            self.update(Configuration.read_file(first_arg))
        else:
            super(Configuration, self).__init__(*args, **kwargs)

    @staticmethod
    def _parse(key):
        if isinstance(key, str):
            return (key,)
        elif key == None or isinstance(key, tuple):
            return key
        else:
            raise KeyError(key)

    def __getitem__(self, key):
        try:
            value = super(Configuration, self).__getitem__(self._parse(key))
        except KeyError:
            value = Section()
            super(Configuration, self).__setitem__(self._parse(key), value)
        return value

    def __setitem__(self, key, value):
        if isinstance(value, Section):
            super(Configuration, self).__setitem__(self._parse(key), value)
        elif isinstance(value, dict):
            super(Configuration, self).__setitem__(self._parse(key), Section(value))
        else:
            raise ValueError('Not section type: %s' % repr(value))

    @classmethod
    def read_input(Class, section=None,
            positional=(), keyword1=(), var_positional=(), keyword2=()):
        '''
        :returns: The configuration
        :type section: Iterable of tuple of str, or None
        :param section: Subcommand to apply the configuration to, or
            None (default) to apply it to all subcommands
        :type positional: Iterable of tuple of str and anything
        :param positional: Positional arguments
        :param keyword1: Keyword arguments that came before a variable
            positional argument
        :type var_positional: Iterable of tuple of str and anything
        :param var_positional: Variable positional arguments (``*args``)
        :param keyword2: Other keyword arguments 
        :rtype: Configuration
        :returns: The configuration
        '''
        merged_inputs = list(positional + keyword1)
        if 1 == len(var_positional):
            name, values = var_positional[0]
            for value in values:
                merged_inputs.append((name, value))
        merged_inputs.extend(keyword2)

        c = Class()
        for key, value in merged_inputs:
            c[section].append(key, value)
        return c

    @classmethod
    def read_file(Class, source):
        '''
        :type source: str or file-like object with "r" mode
        :param source: Configuration file
        :rtype: Configuration
        :returns: The configuration
        '''
        if isinstance(source, str):
            fp = StringIO(source)
        elif hasattr(source, 'read'):
            fp = source
        else:
            raise TypeError('Unrecognized source: %s' % repr(source))
        rows = list(Colon.reader(fp))
        fp.close()
        lengths = set(map(len, rows))

        c = Class()
        if lengths == {2}:
            for argument, value in rows:
                c[None].append(argument, value)
        elif lengths == {3}:
            for section, argument, value in rows:
                c[Space.read(section)].append(argument, value)
        else:
            raise ValueError('Inconsistent formatting in DSV file')
        return c

    @classmethod
    def read_filename(Class, source):
        with open(source) as fp:
            c = Class.read_file(fp)
        return c

    def write(configuration, with_sections=True):
        '''
        :param Configuration configuration: The configuration values to write
        :param bool with_sections: Save different values for arguments of
            the same name in different sections
        :rtype: str
        :returns: Configuration file
        '''
        with StringIO() as fp:
            w = Colon.writer(fp)
            for section, s in configuration.items():
                for argument, values in s.items():
                    if with_sections:
                        left = Space.write(section),
                    else:
                        left = tuple()
                    for value in values:
                        w.writerow(left + (argument, value,))
            text = fp.getvalue()
        return text

    @classmethod
    def merge(Class, *configurations):
        '''
        :param Configuration configurations: The configurations to merge
        '''
        c = Class()
        for configuration in configurations:
            for section, s in configuration.items():
                for argument, value in s.items():
                    c[section][argument] = value
        return c

class ListArgument(list):
    pass

class Section(OrderedDict):
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], Section):
            self.update(args[0])
        else:
            super(Section, self).__init__(*args, **kwargs)

    def flatten(self, defaults={}, list_type=tuple()):
        out = OrderedDict()

        # Apply configurations.
        for key, value in self.items():
            if key in list_type:
                if key not in out:
                    out[key] = ListArgument()
                out[key].extend(value)
            else:
                out[key] = value[-1]

        # Apply defaults.
        for key in defaults:
            if key not in out:
                out[key] = defaults[key]

        return out

    @staticmethod
    def _check(key):
        if not isinstance(key, str):
            raise KeyError('Not str type: %s' % key)

    def __getitem__(self, key):
        self._check(key)
        return super(Section, self).__getitem__(key)

    def __setitem__(self, key, values):
        self._check(key)
        if not isinstance(values, list) and all(isinstance(v, str) for v in values):
            raise TypeError('Argument values must be a list of str')
        return super(Section, self).__setitem__(key, values)

    def append(self, key, value):
        if isinstance(value, str):
            if not key in self:
                self[key] = []
            self[key].append(value)
        else:
            raise TypeError('Argument value must be str, not %s (key %s)' % \
                            (type(value), key))

    def extend(self, key, values):
        for value in values:
            self.append(key, value)



"""
class Inputs(urwid.ListBox):
    def keypress(self, size, key):
        key = super(Inputs, self).keypress(size, key)
        if key == 


def text(program, handlers=None, freedesktop=False):
    '''
    :type program: Program, callable, list, or dict
    :param program: The program for which to produce the interface,
    '''
    import urwid
    urwid.MainLoop(Main).run()

ListBox
    ListBoxRow
    ListBoxRow
    ListBoxRow
        Box
            Button
            Button
    ListBoxRow
       TextView
"""



from distutils.core import setup

packages=[
    'horetu',
    'horetu.interfaces',
    'horetu.interfaces.wsgi',
    'funmap',
    'triedict',
]
extras_require={
  'django': ['Django>=1.9.12'],
  'wsgi': ['WebOb>=1.6.1'],
  'irc': ['irc>=15.0.6'],

  'docs': ['sphinxcontrib-autorun>=0.1'],
  'tests': ['BeautifulSoup4', 'pytest>=2.6.4', 'requests>=2.11.0'],
}
extras_require['all'] = extras_require['wsgi'] + \
                        extras_require['django'] + \
                        extras_require['irc']
extras_require['dev'] = extras_require['all'] + \
                        extras_require['docs'] + \
                        extras_require['tests']

setup(name='horetu',
      author='Thomas Levine',
      author_email='_@thomaslevine.com',
      description='Make user interfaces to functions.',
      url='https://thomaslevine.com/scm/horetu/',
      packages=packages,
      install_requires=[
      ],
      extras_require=extras_require,
      tests_require=[
          'horetu[tests]',
      ],
      classifiers=[
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
      ],
      version='0.4.2',
      license='AGPL',
      )



import sys
import itertools
import textwrap
import shutil
import subprocess
from collections import OrderedDict
from triedict import triedict
from . import _util
from . import _bind
from .. import exceptions
from ..program import Program
from ..annotations import Flag

import os
XDG_OPEN = ('xdg-open', '/dev/stdin')

def cli(program, argv=sys.argv, return_=False,
        handlers=None, freedesktop=False,
        linebreak=True,
        exit_=sys.exit, stdout=sys.stdout.buffer, stderr=sys.stderr.buffer):
    '''
    :type program: Program, callable, list, or dict
    :param program: The program for which to produce the interface
    :param list argv: :py:obj:`sys.argv` by default
    :param bool return_: If this is ``True``, simply return the result of the function;
        do not write stuff to ``exit_``, ``stdout``, or ``stderr``.
    :param function exit_: :py:func:`sys.exit` by default
    :param dict handlers: Mapping from content type to handler program
    :param bool freedesktop: Fall back to xdg-open if a handler is not available
    :param bool linebreak: If the program returns an iterator, horetu will print
        string representations of the elements in the iterator. If this is True,
        horetu will add a newline character after each string element.
    :type stdout: Binary file
    :param stdout: :py:func:`sys.stdout.buffer` by default
    :type stderr: Binary file
    :param stderr: :py:func:`sys.stderr.buffer` by default
    '''
    if not handlers:
        handlers = {}
    argv = list(argv) # Prevent myself from modifying the input.
    if not isinstance(program, Program):
        program = Program(program)
    if not program.name:
        program.name = argv[0]

    def write(stream, fp):
        for line in stream:
            fp.write(line.encode('utf-8') + b'\n')
        fp.flush()

    if return_:
        section, raw_arguments = _bind.web_section(program, argv[1:])
        function = program[section]
        function.require_help()
        xs = from_argv(function, raw_arguments)
        return function.run('raw', *xs)
    else:
        try:
            try:
                section, raw_arguments = _bind.cli_section(program, argv[1:])
                try:
                    function = program[section]
                    function.require_help()
                    xs = from_argv(function, raw_arguments)
                    raw_output = function.run('raw', *xs)

                    rt = function.return_type
                    if rt and (freedesktop or (rt.parsed in handlers)):

                        handler = handlers.get(rt.parsed, XDG_OPEN)

                        if isinstance(raw_output, bytes):
                            stream = [raw_output]
                        else:
                            stream = raw_output
                        p = subprocess.Popen(handler, stdin=subprocess.PIPE, stderr=stderr)
                        for element in stream:
                            if isinstance(element, bytes):
                                p.stdin.write(element)
                            else:
                                raise ValueError('When you set a content type, the function must return either bytes or an iterable of bytes.')
                        p.stdin.close()
                        p.wait()
                    else:
                        suffix = b'\n' if linebreak else b''
                        for line in _util.stream(raw_output):
                            stdout.write(line + suffix)
                            stdout.flush()

                except exceptions.BaseHoretuException as e:
                    e.section = section
                    raise e

            except exceptions.ShowHelp as e:
                write(man(program, e), stdout)
                exit_(0)
            except exceptions.CouldNotParse as e:
                stdout.flush()
                write(usage(program, e), stderr)
                exit_(2)
            except exceptions.Error as e:
                stdout.flush()
                write(usage(program, e), stderr)
                exit_(3)
            else:
                exit_(0)
        except BrokenPipeError:
            stderr.close()

def from_argv(function, raw_arguments):
    mixed = iter(raw_arguments)
    keyword2_params = triedict({param.name: param for param in function.keyword2})

    positional = []
    keyword2 = []
    var_keyword = {}
    before_double_hyphen = True

    while True:
        try:
            raw_arg = next(mixed)
        except StopIteration:
            break

        arg = raw_arg[1:]
        could_not_parse = exceptions.CouldNotParse(
            'Needs parameter -- %s' % raw_arg)
        if before_double_hyphen and raw_arg.startswith('-'):
            if raw_arg == '--':
                before_double_hyphen = False
            elif raw_arg.startswith('--'):
                try:
                    value = next(mixed)
                except StopIteration:
                    raise could_not_parse
                else:
                    var_keyword[raw_arg[2:]] = value
            elif arg in keyword2_params:
                p = keyword2_params[arg]
                if isinstance(p, Flag):
                    value = Flag.YES
                else:
                    try:
                        value = next(mixed)
                    except StopIteration:
                        raise could_not_parse
                keyword2.append((p.name, value))
            else:
                raise exceptions.CouldNotParse('Unknown option -- %s' % arg)
        else:
            positional.append(raw_arg)

    arguments = _bind.positionals(function, positional)
    for name, value in keyword2:
        if not name in arguments:
            arguments[name] = []
        arguments[name].append(value)
    return arguments, var_keyword


def _section_prefix(prog, x):
    for section in sorted(prog, key=len, reverse=True):
        if tuple(section[:len(x)]) == x:
            return x
    return tuple()

def _endpoints(program, section):
    for subsection in program.subset(section):
        s = program[subsection]
        signature = ''
        for x in s.positional:
            signature += ' %s' % x.name
        for x in s.keyword1:
            signature += ' [%s]' % x.name
        if s.var_positional:
            signature += ' [%s ...]' % s.var_positional[0].name
        yield subsection, signature[1:]

def _join(f):
    def decorator(*args, **kwargs):
        return '\n'.join(f(*args, **kwargs))
    return decorator

@_join
def _format_arg(prefix, indent, param):
    columns, _ = shutil.get_terminal_size((80, 20))
    whitespace = ' ' * len(param.name)
    n = columns - len(param.name) - indent - len(prefix) - 2
    
    first = True
    for right in textwrap.wrap(param.description, n):
        if first:
            left = param.name + ': '
            first = False
        else:
            left = whitespace + '  '
        yield (' ' * indent) + left + right

def usage(prog, h):
    p = {
        'name': prog.name,
        'message': h.message,
        'endpoints': _endpoints(prog, h.section if h.section in prog else tuple()),
    }
    if p['message']:
        yield 'error: %(message)s' % p

    for i, (section, signature) in enumerate(p['endpoints']):
        yield _usage_line(prog, section, signature,
                          'usage: ' if i == 0 else '       ')

def _usage_line(program, section, signature, prefix):
    function = program[section]
    function.require_help()
    help_flag = function.help_flag()
    q = {
        'prefix': prefix,
        'name': program.name,
        'sub': (' ' + ' '.join(section)).rstrip() + ' ',
        'signature': signature,
        'sep': ' [--] ' if any(signature) else ' ',
        'help': ('[-%s] ' % help_flag) if help_flag else '',
    }
    return '%(prefix)s%(name)s%(sub)s%(help)s[options]%(sep)s%(signature)s' % q

def man(prog, h):
    columns, _ = shutil.get_terminal_size((80, 20))
    f = prog[h.section]
    p =  {
        'name': prog.name,
        'endpoints': _endpoints(prog, h.section),
        'description': f.description,
        'args': (_format_arg(' ', 2, a) for a in f.all_positionals()),
        'kwargs': (_format_arg('-', 2, a) for a in f.keyword2),
    }

    yield 'SYNOPSIS'
    for section, signature in p['endpoints']:
        yield _usage_line(prog, section, signature, '  ')
    if p['description']:
        yield 'DESCRIPTION'
        for line in textwrap.wrap(p['description'], columns-2):
            yield '  ' + line
    if p['args']:
        yield 'INPUTS'
        for arg in p['args']:
            yield arg
        try:
            arg
        except NameError:
            yield '  (None)'
    yield 'OPTIONS'
    for kwarg in p['kwargs']:
        yield kwarg



import re
from functools import wraps

def ensure_bytes(x):
    if x == None:
        raise StopIteration
    elif isinstance(x, bytes):
        yield x
        raise StopIteration
    elif hasattr(x, '__iter__') and not isinstance(x, str):
        for y in x:
            if isinstance(y, bytes):
                yield y
            else:
                break
        else:
            raise StopIteration
    raise TypeError('Functions in this interface must return bytes or iterables of bytes.')

def _url_template(section, s):
    signature = ''
    for x in section:
        signature += '/%s' % x
    for x in s.positional:
        signature += '/:%s' % x.name
    for x in s.keyword1:
        signature += '/[%s]' % x.name
    for x in s.var_positional:
        signature += '/&lt;%s&gt;' % x.name
    signature += '?(option=...)'
    return signature

def _as_bytes(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        text = '\n'.join(function(*args, **kwargs))
        return ('<main>%s</main>' % text).encode('utf-8')
    return wrapper

def _endpoints(prog, section):
    for subsection in prog.subset(section):
        function = prog[subsection]
        yield (
            '/' + '/'.join(subsection),
            _url_template(subsection, function),
        )

@_as_bytes
def usage(prog, h):
    endpoints = _endpoints(prog, h.section)

    if h.message:
        yield '<em>%s</em>' % h.message
    yield '<ul>'
    for endpoint in endpoints:
        yield '<li><a href="%s"><code>%s</code></a></li>' % endpoint
    yield '</ul>'

@_as_bytes
def man(prog, h):
    function = prog[h.section] # XXX
    endpoints = _endpoints(prog, h.section)
    description = list(filter(None, re.split('r[\n\r]{2,}', function.description)))
    args = function.all_positionals()
    kwargs = function.keyword2

    yield '''\
<h2>Synopsis<h2>
<ul>
'''
    for i, signature in enumerate(endpoints):
        yield '<li><a href="%s"><code>%s</code></a></li>' % signature
    yield '</ul>'
    if description:
        yield '<h2>Description</h2>'
        for p in description:
            yield '<p>%s</p>' % p

    if args:
        yield '''\
<h3>Inputs</h3>
<dl>
'''
        for arg in args:
            yield '<dt>%s</dt><dd>%s</dt>\n' % (arg.name, arg.description)
        yield '</dl>\n'

    if kwargs:
        yield '''\
<h3>Options</h3>
<dl>
'''
        for pair in kwargs:
            yield '<dt>%s</dt><dd>%s</dd>' % pair
        yield '</dl>'

    yield '''\
<h2>Detail</h2>
<p>
  Put "help" in the query string ("?help") for more help.
</p>
'''



import shlex
from itertools import accumulate
from functools import partial
from collections import OrderedDict
from ..exceptions import CouldNotParse, Exit
from ..annotations import Flag

def positionals(function, raw_values):
    values = list(raw_values)
    positional = []
    keyword1 = []
    var_positional = []

    while 0 < len(values):
        if len(positional) < len(function.positional):
            name = function.positional[len(positional)].name
            value = values.pop(0)
            positional.append((name, value))
        elif len(keyword1) < len(function.keyword1):
            name = function.keyword1[len(keyword1)].name
            value = values.pop(0)
            keyword1.append((name, value))
        elif len(var_positional) == 0 and len(function.var_positional) == 1:
            name = function.var_positional[0].name
            for value in values:
                var_positional.append((name, value))
            values.clear()
        else:
            raise CouldNotParse('Extra positional arguments: %s' % _sh(values))

    arguments = OrderedDict()
    for name, value in positional + keyword1 + var_positional:
        if not name in arguments:
            arguments[name] = []
        arguments[name].append(value)

    return arguments

def split_section(allow, program, xs):
    for section in sorted(program, key=len, reverse=True):
        if tuple(xs[:len(section)]) == section:
            valid = True
            break
    else:
        valid = False
    
    if valid and section in program:
        return section, xs[len(section):]
    else:
        for section in accumulate((a,) for a in xs if allow(a)):
            if section not in program:
                break
        else:
            section = tuple()
        if xs:
            msg = 'Bad subcommand'
        else:
            msg = 'Needs subcommand'
        raise CouldNotParse(section=section, message=msg)


def _sh(values):
    return ' '.join(map(shlex.quote, values))

def _yes_hyphen(x):
    return x.startswith('-')
def _no_hyphen(x):
    return not x.startswith('-')

cli_section = partial(split_section, _no_hyphen)
web_section = partial(split_section, lambda x: True)
